! WHIZARD 2.2.0 alpha Sep 12 2013
! 
! Copyright (C) 1999-2013 by 
!     Wolfgang Kilian <kilian@physik.uni-siegen.de>
!     Thorsten Ohl <ohl@physik.uni-wuerzburg.de>
!     Juergen Reuter <juergen.reuter@desy.de>
!     
!     with contributions by 
!     Christian Speckner <cnspeckn@googlemail.com> 
!     and  Felix Braam, Sebastian Schmidt, Daniel Wiesler 
!
! WHIZARD is free software; you can redistribute it and/or modify it
! under the terms of the GNU General Public License as published by 
! the Free Software Foundation; either version 2, or (at your option)
! any later version.
!
! WHIZARD is distributed in the hope that it will be useful, but
! WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
! GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this program; if not, write to the Free Software
! Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! This file has been stripped of most comments.  For documentation, refer
! to the source 'whizard.nw'

module sf_mappings

  use kinds, only: default !NODEP!
  use file_utils !NODEP!
  use diagnostics !NODEP!
  use unit_tests

  implicit none
  private

  public :: sf_mapping_t
  public :: sf_s_mapping_t
  public :: sf_channel_t
  public :: allocate_sf_channels
  public :: any_sf_channel_has_mapping
  public :: sf_mappings_test

  integer, parameter :: SFMAP_NONE = 0
  integer, parameter :: SFMAP_SINGLE = 1
  integer, parameter :: SFMAP_MULTI_S = 2
  integer, parameter :: SFMAP_MULTI_RES = 3
  integer, parameter :: SFMAP_MULTI_ONS = 4


  type, abstract :: sf_mapping_t
     integer, dimension(:), allocatable :: i
   contains
     procedure (sf_mapping_write), deferred :: write
     procedure :: base_init => sf_mapping_base_init
     procedure (sf_mapping_compute), deferred :: compute
     procedure (sf_mapping_inverse), deferred :: inverse
     procedure :: integral => sf_mapping_integral
  end type sf_mapping_t
     
  type, extends (sf_mapping_t) :: sf_s_mapping_t
     logical :: power_set = .false.
     real(default) :: power = 1._default
   contains
     procedure :: write => sf_s_mapping_write
     procedure :: init => sf_s_mapping_init
     procedure :: compute => sf_s_mapping_compute
     procedure :: inverse => sf_s_mapping_inverse
  end type sf_s_mapping_t
  
  type :: sf_channel_t
     integer, dimension(:), allocatable :: map_code
     class(sf_mapping_t), allocatable :: multi_mapping
   contains
     procedure :: write => sf_channel_write
     procedure :: init => sf_channel_init
     generic :: assignment (=) => sf_channel_assign
     procedure :: sf_channel_assign
     procedure :: activate_mapping => sf_channel_activate_mapping
     procedure :: set_s_mapping => sf_channel_set_s_mapping
     procedure :: is_single_mapping => sf_channel_is_single_mapping
     procedure :: is_multi_mapping => sf_channel_is_multi_mapping
     procedure :: set_par_index => sf_channel_set_par_index
  end type sf_channel_t
  

  abstract interface
     subroutine sf_mapping_write (object, unit)
       import
       class(sf_mapping_t), intent(in) :: object
       integer, intent(in), optional :: unit
     end subroutine sf_mapping_write
  end interface

  abstract interface
     subroutine sf_mapping_compute (mapping, r, f, p)
       import
       class(sf_mapping_t), intent(in) :: mapping
       real(default), dimension(:), intent(out) :: r
       real(default), intent(out) :: f
       real(default), dimension(:), intent(in) :: p
     end subroutine sf_mapping_compute
  end interface
  
  abstract interface
     subroutine sf_mapping_inverse (mapping, r, f, p)
       import
       class(sf_mapping_t), intent(in) :: mapping
       real(default), dimension(:), intent(in) :: r
       real(default), intent(out) :: f
       real(default), dimension(:), intent(out) :: p
     end subroutine sf_mapping_inverse
  end interface
  

contains

  subroutine sf_mapping_base_init (mapping, n_par)
    class(sf_mapping_t), intent(out) :: mapping
    integer, intent(in) :: n_par
    allocate (mapping%i (n_par))
    mapping%i = 0
  end subroutine sf_mapping_base_init
    
  function sf_mapping_integral (mapping, n_calls) result (integral)
    class(sf_mapping_t), intent(in) :: mapping
    integer, intent(in) :: n_calls
    real(default) :: integral
    integer :: n_bin, i, j
    real(default), dimension(2) :: p, r
    real(default) :: dx, f, s
    n_bin = nint (sqrt (real (n_calls)))
    dx = 1._default / n_bin
    s = 0
    do i = 1, n_bin
       p(1) = i * dx - dx/2
       do j = 1, n_bin
          p(2) = j * dx - dx/2
          call mapping%compute (r, f, p)
          s = s + f
       end do
    end do
    integral = s / n_bin / n_bin
  end function sf_mapping_integral

  subroutine sf_s_mapping_write (object, unit)
    class(sf_s_mapping_t), intent(in) :: object
    integer, intent(in), optional :: unit
    integer :: u
    u = output_unit (unit)
    write (u, "(1x,A)", advance="no")  "map"
    if (any (object%i /= 0)) then
       write (u, "('(',I0,',',I0,')')", advance="no")  object%i
    end if
    write (u, "(A,F7.5,A)")  ": standard (", object%power, ")"
  end subroutine sf_s_mapping_write
  
  subroutine sf_s_mapping_init (mapping, power)
    class(sf_s_mapping_t), intent(out) :: mapping
    real(default), intent(in), optional :: power
    call mapping%base_init (2)
    if (present (power)) then
       mapping%power_set = .true.
       mapping%power = power
    end if
  end subroutine sf_s_mapping_init
    
  subroutine sf_s_mapping_compute (mapping, r, f, p)
    class(sf_s_mapping_t), intent(in) :: mapping
    real(default), dimension(:), intent(out) :: r
    real(default), intent(out) :: f
    real(default), dimension(:), intent(in) :: p
    real(default), dimension(2) :: r2
    integer :: j
    if (mapping%power_set) then
       call map_unit_square (r2, f, p(mapping%i), mapping%power)
    else
       call map_unit_square (r2, f, p(mapping%i))
    end if
    do j = 1, 2
       r(mapping%i(j)) = r2(j)
    end do
  end subroutine sf_s_mapping_compute

  subroutine sf_s_mapping_inverse (mapping, r, f, p)
    class(sf_s_mapping_t), intent(in) :: mapping
    real(default), dimension(:), intent(in) :: r
    real(default), intent(out) :: f
    real(default), dimension(:), intent(out) :: p
    real(default), dimension(2) :: p2
    integer :: j
    if (mapping%power_set) then
       call map_unit_square_inverse (r(mapping%i), f, p2, mapping%power)
    else
       call map_unit_square_inverse (r(mapping%i), f, p2)
    end if
    do j = 1, 2
       p(mapping%i(j)) = p2(j)
    end do
  end subroutine sf_s_mapping_inverse

  subroutine map_unit_square (r, factor, p, power)
    real(default), dimension(2), intent(out) :: r
    real(default), intent(out) :: factor
    real(default), dimension(2), intent(in) :: p
    real(default), intent(in), optional :: power
    real(default) :: xx, yy
    factor = 1
    xx = p(1)
    yy = p(2)
    if (present(power)) then
       if (p(1) > 0 .and. power > 1) then
          xx = p(1)**power
          factor = factor * power * xx / p(1)
       end if
    end if
    if (xx /= 0) then
       r(1) = xx ** yy
       r(2) = xx / r(1)
       factor = factor * abs (log (xx))
    else
       r = 0
    end if
  end subroutine map_unit_square

  subroutine map_unit_square_inverse (r, factor, p, power)
    real(kind=default), dimension(2), intent(in) :: r
    real(kind=default), intent(out) :: factor
    real(kind=default), dimension(2), intent(out) :: p
    real(kind=default), intent(in), optional :: power
    real(kind=default) :: lg, xx, yy
    factor = 1
    xx = r(1) * r(2)
    if (xx /= 0) then
       lg = log (xx)
       yy = log (r(1)) / lg
       p(2) = yy
       factor = factor * abs (lg)
       if (present(power)) then
          p(1) = xx**(1._default/power)
          factor = factor * power * xx / p(1)
       else
          p(1) = xx
       end if
    else
       p = 0
    end if
  end subroutine map_unit_square_inverse

  subroutine sf_channel_write (object, unit)
    class(sf_channel_t), intent(in) :: object
    integer, intent(in), optional :: unit
    integer :: u, i
    u = output_unit (unit)
    if (allocated (object%map_code)) then
       do i = 1, size (object%map_code)
          select case (object%map_code (i))
          case (SFMAP_NONE)
             write (u, "(1x,A)", advance="no") "-"
          case (SFMAP_SINGLE)
             write (u, "(1x,A)", advance="no") "+"
          case (SFMAP_MULTI_S)
             write (u, "(1x,A)", advance="no") "s"
          case (SFMAP_MULTI_RES)
             write (u, "(1x,A)", advance="no") "r"
          case (SFMAP_MULTI_ONS)
             write (u, "(1x,A)", advance="no") "o"
          end select
       end do
    else
       write (u, "(1x,A)", advance="no") "-"
    end if
    if (allocated (object%multi_mapping)) then
       write (u, "(1x,'/')", advance="no")
       call object%multi_mapping%write (u)
    else
       write (u, *)
    end if
  end subroutine sf_channel_write
       
  subroutine sf_channel_init (channel, n_strfun)
    class(sf_channel_t), intent(out) :: channel
    integer, intent(in) :: n_strfun
    allocate (channel%map_code (n_strfun))
    channel%map_code = SFMAP_NONE
  end subroutine sf_channel_init
    
  subroutine sf_channel_assign (copy, original)
    class(sf_channel_t), intent(out) :: copy
    type(sf_channel_t), intent(in) :: original
    allocate (copy%map_code (size (original%map_code)))
    copy%map_code = original%map_code
    if (allocated (original%multi_mapping)) then
       allocate (copy%multi_mapping, source = original%multi_mapping)
    end if
  end subroutine sf_channel_assign
  
  subroutine allocate_sf_channels (channel, n_channel, n_strfun)
    type(sf_channel_t), dimension(:), intent(out), allocatable :: channel
    integer, intent(in) :: n_channel
    integer, intent(in) :: n_strfun
    integer :: c
    allocate (channel (n_channel))
    do c = 1, n_channel
       call channel(c)%init (n_strfun)
    end do
  end subroutine allocate_sf_channels

  subroutine sf_channel_activate_mapping (channel, i_sf)
    class(sf_channel_t), intent(inout) :: channel
    integer, dimension(:), intent(in) :: i_sf
    channel%map_code(i_sf) = SFMAP_SINGLE
  end subroutine sf_channel_activate_mapping
  
  subroutine sf_channel_set_s_mapping (channel, i_sf, power)
    class(sf_channel_t), intent(inout) :: channel
    integer, dimension(:), intent(in) :: i_sf
    real(default), intent(in), optional :: power
    channel%map_code(i_sf) = SFMAP_MULTI_S
    allocate (sf_s_mapping_t :: channel%multi_mapping)
    select type (mapping => channel%multi_mapping)
    type is (sf_s_mapping_t)
       call mapping%init (power)
    end select
  end subroutine sf_channel_set_s_mapping

  function sf_channel_is_single_mapping (channel, i_sf) result (flag)
    class(sf_channel_t), intent(in) :: channel
    integer, intent(in) :: i_sf
    logical :: flag
    flag = channel%map_code(i_sf) == SFMAP_SINGLE
  end function sf_channel_is_single_mapping 

  function sf_channel_is_multi_mapping (channel, i_sf) result (flag)
    class(sf_channel_t), intent(in) :: channel
    integer, intent(in) :: i_sf
    logical :: flag
    select case (channel%map_code(i_sf))
    case (SFMAP_NONE, SFMAP_SINGLE)
       flag = .false.
    case default
       flag = .true.
    end select
  end function sf_channel_is_multi_mapping 

  function any_sf_channel_has_mapping (channel) result (flag)
    type(sf_channel_t), dimension(:), intent(in) :: channel
    logical :: flag
    integer :: c
    flag = .false.
    do c = 1, size (channel)
       flag = flag .or. any (channel(c)%map_code /= SFMAP_NONE)
    end do
  end function any_sf_channel_has_mapping
  
  subroutine sf_channel_set_par_index (channel, j, i_par)
    class(sf_channel_t), intent(inout) :: channel
    integer, intent(in) :: j
    integer, intent(in) :: i_par
    channel%multi_mapping%i(j) = i_par
  end subroutine sf_channel_set_par_index
    

  subroutine sf_mappings_test (u, results)
    integer, intent(in) :: u
    type(test_results_t), intent(inout) :: results
    call test (sf_mappings_1, "sf_mappings_1", &
         "standard pair mapping", &
         u, results)
    call test (sf_mappings_2, "sf_mappings_2", &
         "structure-function mapping channels", &
         u, results)
  end subroutine sf_mappings_test
  
  subroutine sf_mappings_1 (u)
    integer, intent(in) :: u
    class(sf_mapping_t), allocatable :: mapping
    real(default), dimension(2) :: p, r
    real(default) :: f
    
    write (u, "(A)")  "* Test output: sf_mappings_1"
    write (u, "(A)")  "*   Purpose: probe standard mapping"
    write (u, "(A)")
    
    allocate (sf_s_mapping_t :: mapping)
    select type (mapping)
    type is (sf_s_mapping_t)
       call mapping%init ()
       mapping%i(1) = 1
       mapping%i(2) = 2
    end select
       
    call mapping%write (u)
    
    write (u, *)
    write (u, "(A)")  "Probe at (0,0):"
    p = [0._default, 0._default]
    
    call mapping%compute (r, f, p)
    write (u, "(3x,A,2(1x,F7.5))")  "p =", p
    write (u, "(3x,A,2(1x,F7.5))")  "r =", r
    write (u, "(3x,A,2(1x,F7.5))")  "f =", f

    write (u, *)
    call mapping%inverse (r, f, p)
    write (u, "(3x,A,2(1x,F7.5))")  "p =", p
    write (u, "(3x,A,2(1x,F7.5))")  "r =", r
    write (u, "(3x,A,2(1x,F7.5))")  "f =", f

    write (u, *)
    write (u, "(A)")  "Probe at (0.5,0.5):"
    p = [0.5_default, 0.5_default]
    
    call mapping%compute (r, f, p)
    write (u, "(3x,A,2(1x,F7.5))")  "p =", p
    write (u, "(3x,A,2(1x,F7.5))")  "r =", r
    write (u, "(3x,A,2(1x,F7.5))")  "f =", f

    write (u, *)
    call mapping%inverse (r, f, p)
    write (u, "(3x,A,2(1x,F7.5))")  "p =", p
    write (u, "(3x,A,2(1x,F7.5))")  "r =", r
    write (u, "(3x,A,2(1x,F7.5))")  "f =", f

    write (u, *)
    write (u, "(A)")  "Probe at (0.1,0.5):"
    p = [0.1_default, 0.5_default]
    
    call mapping%compute (r, f, p)
    write (u, "(3x,A,2(1x,F7.5))")  "p =", p
    write (u, "(3x,A,2(1x,F7.5))")  "r =", r
    write (u, "(3x,A,2(1x,F7.5))")  "f =", f

    write (u, *)
    write (u, "(A)")  "Probe at (0.1,0.1):"
    p = [0.1_default, 0.1_default]
    
    call mapping%compute (r, f, p)
    write (u, "(3x,A,2(1x,F7.5))")  "p =", p
    write (u, "(3x,A,2(1x,F7.5))")  "r =", r
    write (u, "(3x,A,2(1x,F7.5))")  "f =", f

    write (u, *)
    write (u, "(A)")  "Compute integral:"
    write (u, "(3x,A,1x,F7.5)") "I =", mapping%integral (100000)

    deallocate (mapping)
    allocate (sf_s_mapping_t :: mapping)
    select type (mapping)
    type is (sf_s_mapping_t)
       call mapping%init (power=2._default)
       mapping%i(1) = 1
       mapping%i(2) = 2
    end select
       
    write (u, *)
    call mapping%write (u)
    
    write (u, *)
    write (u, "(A)")  "Probe at (0,0):"
    p = [0._default, 0._default]
    
    call mapping%compute (r, f, p)
    write (u, "(3x,A,2(1x,F7.5))")  "p =", p
    write (u, "(3x,A,2(1x,F7.5))")  "r =", r
    write (u, "(3x,A,2(1x,F7.5))")  "f =", f

    write (u, *)
    call mapping%inverse (r, f, p)
    write (u, "(3x,A,2(1x,F7.5))")  "p =", p
    write (u, "(3x,A,2(1x,F7.5))")  "r =", r
    write (u, "(3x,A,2(1x,F7.5))")  "f =", f

    write (u, *)
    write (u, "(A)")  "Probe at (0.5,0.5):"
    p = [0.5_default, 0.5_default]
    
    call mapping%compute (r, f, p)
    write (u, "(3x,A,2(1x,F7.5))")  "p =", p
    write (u, "(3x,A,2(1x,F7.5))")  "r =", r
    write (u, "(3x,A,2(1x,F7.5))")  "f =", f

    write (u, *)
    call mapping%inverse (r, f, p)
    write (u, "(3x,A,2(1x,F7.5))")  "p =", p
    write (u, "(3x,A,2(1x,F7.5))")  "r =", r
    write (u, "(3x,A,2(1x,F7.5))")  "f =", f

    write (u, *)
    write (u, "(A)")  "Probe at (0.1,0.5):"
    p = [0.1_default, 0.5_default]
    
    call mapping%compute (r, f, p)
    write (u, "(3x,A,2(1x,F7.5))")  "p =", p
    write (u, "(3x,A,2(1x,F7.5))")  "r =", r
    write (u, "(3x,A,2(1x,F7.5))")  "f =", f

    write (u, *)
    write (u, "(A)")  "Probe at (0.1,0.1):"
    p = [0.1_default, 0.1_default]
    
    call mapping%compute (r, f, p)
    write (u, "(3x,A,2(1x,F7.5))")  "p =", p
    write (u, "(3x,A,2(1x,F7.5))")  "r =", r
    write (u, "(3x,A,2(1x,F7.5))")  "f =", f

    write (u, *)
    write (u, "(A)")  "Compute integral:"
    write (u, "(3x,A,1x,F7.5)")  "I =", mapping%integral (100000)

    write (u, "(A)")
    write (u, "(A)")  "* Test output end: sf_mappings_1"

  end subroutine sf_mappings_1
  
  subroutine sf_mappings_2 (u)
    integer, intent(in) :: u
    type(sf_channel_t), dimension(:), allocatable :: channel
    integer :: c
    
    write (u, "(A)")  "* Test output: sf_mappings_2"
    write (u, "(A)")  "*   Purpose: construct and display &
         &mapping-channel objects"
    write (u, "(A)")
    
    call allocate_sf_channels (channel, n_channel = 4, n_strfun = 2)
    call channel(2)%activate_mapping ([1])
    call channel(3)%set_s_mapping ([1,2])
    call channel(4)%set_s_mapping ([1,2], power=2._default)
    
    call channel(3)%set_par_index (1, 1)
    call channel(3)%set_par_index (2, 4)

    call channel(4)%set_par_index (1, 1)
    call channel(4)%set_par_index (2, 4)

    do c = 1, size (channel)
       write (u, "(I0,':')", advance="no")  c
       call channel(c)%write (u)
    end do

    write (u, "(A)")
    write (u, "(A)")  "* Test output end: sf_mappings_2"

  end subroutine sf_mappings_2
  

end module sf_mappings
